// ExtrapBC.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, March 8, 2000
#ifndef _EXTRAPBC_H_
#define _EXTRAPBC_H_

#include "Box.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "LoHiSide.H"
#include "GhostBC.H"
/// Class to enforce homogeneous neumann    boundary conditions 
/**
   BoxGhostBC class implentation to enforce homogeneous neumann 
   boundary conditions 
   Recall that
   BoxGhostBC is a class to encapsulate the 
   operations  of ghost-cell boundary conditions at a face.  
   If the solution is phi and the face normal direction is x,
   the boundary conditions usually used can be expressed as
   A*phi + B*dphi/dx = C.
   For homogeneous Neumann BCs, A=C=0, B= 1
   These functions are only meant to be called by DomainGhostBC

 */
class ExtrapBC : public BoxGhostBC
{
public:
  friend class DomainGhostBC;

  ///
  ~ExtrapBC();

  ///
  ExtrapBC();
  
  ///
  ExtrapBC(int dir, Side::LoHiSide sd);

  ///
  ExtrapBC(int dir, Side::LoHiSide sd, 
	    const Interval& a_comps);

protected:

  ///
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
               FArrayBox& a_dircfac,
               FArrayBox& a_inhmval,
               Real a_dx,
               const ProblemDomain& a_domain) const;

  ///
  virtual void applyHomogeneousBCs(FArrayBox& a_state,
				   const ProblemDomain& a_domain,
				   Real a_dx) const;
  
  ///
  virtual void applyInhomogeneousBCs(FArrayBox& a_state,
				     const ProblemDomain& a_domain,
				     Real a_dx) const;
  
  ///
  virtual void applyExtrapBCs(FArrayBox& a_state, 
			      const ProblemDomain& a_domain,
			      Real a_dx) const;


  /// deprecated interface
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
               FArrayBox& a_dircfac,
               FArrayBox& a_inhmval,
               Real a_dx,
               const Box& a_domain) const;

  /// deprecated interface
  virtual void applyHomogeneousBCs(FArrayBox& a_state,
				   const Box& a_domain,
				   Real a_dx) const;
  
  /// deprecated interface
  virtual void applyInhomogeneousBCs(FArrayBox& a_state,
				     const Box& a_domain,
				     Real a_dx) const;
  
  /// deprecated interface
  virtual void applyExtrapBCs(FArrayBox& a_state, 
			      const Box& a_domain,
			      Real a_dx) const;


  ///
  BoxGhostBC* new_boxghostbc() const;

private:
  ExtrapBC(const ExtrapBC& neuin){;}
  void operator=(const ExtrapBC& neuin){;}

};



#endif
