/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// scalarDirichletBC.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#ifndef _scalarDirichletBC_H_
#define _scalarDirichletBC_H_

#include "Box.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "LoHiSide.H"
#include "GhostBC.H"


/// implements dirichlet BCs for scalars
class scalarDirichletBC  : public BoxGhostBC
{

public:
  friend class DomainGhostBC;

  /// 
  scalarDirichletBC(int dir, Side::LoHiSide sd);
  
  /// 
  scalarDirichletBC(int dir, Side::LoHiSide sd, 
		    const Interval& a_comps);

  ///
  virtual ~scalarDirichletBC();

  ///
  scalarDirichletBC();
  
  /// sets Dirichlet Boundary value
  void setVal(const Real a_bcVal);


private:
  
  /// dirichlet value at edge
  Real m_BCVal;
  
  /// fills dirichletfactor with 1.0 and inhomVal with BCVal
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dirfac,
	       FArrayBox& a_inhmval,
	       Real dx, 
	       const ProblemDomain& a_domain) const;
  
  /// fills dirichletfactor with 1.0 and inhomVal with BCVal
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dirfac,
	       FArrayBox& a_inhmval,
	       Real dx, 
	       const Box& a_domain) const;

  ///
  virtual BoxGhostBC* new_boxghostbc() const;

private:
  scalarDirichletBC(const scalarDirichletBC& bcin) {;}

};

///
class scalarInflowBC : public BoxGhostBC
{

public: 
  friend class DomainGhostBC;

  ///
  scalarInflowBC();

  /// 
  scalarInflowBC(int dir, Side::LoHiSide sd);

  ///
  scalarInflowBC(int dir, Side::LoHiSide sd,
		 const Interval& a_comps);


  ///
  virtual ~scalarInflowBC();

  /// set which scalar this is
  void setScalarType(int a_scalType);
  
  ///
  int scalarType() const;

  /// 
  void isHomogeneous(bool a_isHomogeneous);
  
  /// 
  bool isHomogeneous() const;



private:

  ///
  int m_scalType;

  /// 
  bool m_isHomogeneous;

  /// fills dirichlet factor with 1.0 and inhomval with inflow profile
  virtual void 
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dirfac,
	       FArrayBox& a_inhmval,
	       Real dx,
	       const ProblemDomain& a_domain) const;


  /// fills dirichlet factor with 1.0 and inhomval with inflow profile
  virtual void 
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dirfac,
	       FArrayBox& a_inhmval,
	       Real dx,
	       const Box& a_domain) const;


  ///
  virtual BoxGhostBC* new_boxghostbc() const;


};

#endif


