/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AmrNavierStokesFactory.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/
#include "AMRNavierStokesFactory.H"

AMRNavierStokesFactory::AMRNavierStokesFactory() 
{
  setDefaultValues();
}

AMRNavierStokesFactory::~AMRNavierStokesFactory() 
{
  if (m_physBCPtr != NULL) {
    delete m_physBCPtr;
    m_physBCPtr = NULL;
  }
}

void
AMRNavierStokesFactory::CFL(const Real a_cfl) 
{
  m_cfl = a_cfl;
}

void
AMRNavierStokesFactory::refinementThreshold(const Real a_refine_threshold)
{
  m_refine_thresh = a_refine_threshold;
}

void
AMRNavierStokesFactory::setPhysBC(const PhysBCUtil& a_physBC)
{
  if (m_physBCPtr != NULL) {
    delete m_physBCPtr;
    m_physBCPtr = NULL;
  }
  m_physBCPtr = a_physBC.newPhysBCUtil();
}


// "virtual constructor"
AMRLevel*
AMRNavierStokesFactory::new_amrlevel() const
{
  AMRNavierStokes* amrns_ptr = new AMRNavierStokes();
  amrns_ptr->CFL(m_cfl);
  amrns_ptr->refinementThreshold(m_refine_thresh);
  assert (m_physBCPtr != NULL);
  amrns_ptr->setPhysBC(*m_physBCPtr);
 
  return (static_cast <AMRLevel*> (amrns_ptr));
}


void
AMRNavierStokesFactory::setDefaultValues()
{
  m_cfl = 0.5;
  // don't really know what default value to put here
  m_refine_thresh = -1.0;
  // set default Ptr?
  m_physBCPtr = new PhysBCUtil;
  
}


