// Timer.cpp

#ifdef TIMER 

#include <iostream>
#include <fstream>
#include <set>
#include <vector>
#include <cstdio>
#include <cassert>
using namespace std;

#include "parstream.H"
#include "Timer.H"

//#define MEMORY_USAGE

#ifdef MEMORY_USAGE
#include "memusage.H"
char stuff[180];
#endif

// Must initialize the static list defined in Timer.hh
list<Timer*> *Timer::TimerList = NULL; 
//#define NQ 1000000
//extern unsigned long int qiter;
//extern unsigned int qsubbox_len0[NQ];
//extern unsigned int qsubbox_len1[NQ];
//extern unsigned int qsubbox_len2[NQ];
//extern unsigned int qNN[NQ];


static int ID_counter = 0;
static double NtotalStartStops = 0;

#ifdef PAPI
static float CPU_MHZ = 0.1;

// These are just helper functions that call PAPI functions
//   Only used here and only for making code easier to read.
static int PAPI_initSingleEvent(const int event, int *tag);
static int PAPI_initDualEvent(const int event1, const int event2, int *tag);
static int PAPI_initMultiEvent(const int event1, const int event2, 
				const int event3, int *tag);
static int PAPI_checkEvent(const int event);

static int PAPI_init(void);
static void PAPI_hwinfo(float &mhz);
static void PAPI_initMultiplex(void);
static void PAPI_cleanup(int *i);
double computeDerivedCounter(double pc1, double pc2);
#endif

// These are helper functions to write the formatted output
//  summary.  Nothing special.
static void writeDoubleLineSeparator(FILE *out);
static void writeSingleLineSeparator(FILE *out);
static void writeTableHeader(FILE *out);
static void writeLineOfData(FILE *out,
			    const char *name,
			    //const long int count,
                            const double count,
			    const double percent,
			    const double wc_avg,
			    const double wc_min,
			    const double wc_max,
			    const double counter1,
			    const double counter2,
                            const int number_procs);

static void writeTableTotals(FILE *out, 
                             const double table_count_sum,
			     const double parent_avg,
			     const double table_avg_sum, 
			     const double table_min_sum,
			     const double table_max_sum,
                             const double table_pc1_sum,
                             const double table_pc2_sum,
                             const double TimerCost);

// Timer construction for the root of a tree. 
Timer::Timer(const string& name):
  timer_name(name),  Parent(*this)  {
   
  //printf(" Root Timer: %s  ID_counter=%d\n", Name().c_str(), ID_counter);
  setup();

  // Make a list of pointers of all Managed Timers.
  if(TimerList==NULL) TimerList = new list<Timer*>();
  TimerList->push_back(this);  
}

// Non-root Managed Timer construction
Timer::Timer(const string& name, Timer& parent):
  timer_name(name),  Parent(parent) {

  //printf(" Timer: %s  ID_counter=%d\n", Name().c_str(), ID_counter);
  setup();

  if(TimerList==NULL) TimerList = new list<Timer*>();
  TimerList->push_back(this);  
}

// Diagnostic Non-root Managed Timer construction for separate table
Timer::Timer(const string& name, Timer& parent, const int table):
  timer_name(name),  Parent(parent) {

  //printf(" Timer: %s  ID_counter=%d\n", Name().c_str(), ID_counter);
  setup();
  diagnostic_table = table;

  if(TimerList==NULL) TimerList = new list<Timer*>();
  TimerList->push_back(this);  
}

// // Counter Timer
// Timer::Timer(const string& name, const int table):
//   timer_name(name),  Parent(parent) {

//   //printf(" Timer: %s  ID_counter=%d\n", Name().c_str(), ID_counter);
//   setup();
//   diagnostic_table = table;

//   if(TimerList==NULL) TimerList = new list<Timer*>();
//   TimerList->push_back(this);  
// }


// Unmanaged Timer construction
Timer::Timer(): Parent(*this) {
  setup();
}


void Timer::setup() {

  count = 0;
  diagnostic_table = 0;
  //timer_on = false;

  //printf(" Timer: %s  ID_counter=%d\n", Name().c_str(), ID_counter);

  accumulated_WCtime = 0.0;
  ID = ID_counter;
  //if(ID_counter==0) cout << " First Timer Initialized " << endl;
  ID_counter++;

#ifdef PAPI
  //printf(" ID=%d\n", ID);

  // this only needs to be called once. 
  // and before any PAPI calls.  not sure where to put it.
  if(CPU_MHZ < 1.0) {
    PAPI_init();
    PAPI_hwinfo(CPU_MHZ);
  }
  
  //PAPI_initMultiplex();

  accumulated_counter1 = 0;
  accumulated_counter2 = 0;

  values[0] = 0;
  values[1] = 0;

  //PAPI_initSingleEvent(PAPI_TOT_CYC, &myEventSet);

  if(TIMER_COUNTER == 0) {
    PAPI_initDualEvent(PAPI_TOT_CYC, PAPI_FP_INS, &ID);
  } else if(TIMER_COUNTER == 1) {
    PAPI_initDualEvent(PAPI_L1_TCM,  PAPI_L2_TCM, &ID);
  } else {
    PAPI_initDualEvent(PAPI_TOT_INS, PAPI_BR_INS, &ID);
  }
  
  //PAPI_initMultiplex();
  //printf(" initting Timer %d\n", ID);
  //PAPI_initMultiEvent(PAPI_FP_INS, PAPI_TOT_CYC, PAPI_L1_TCM, &ID);

#endif

}

// Destructor for Managed and Unmanaged Timers.
Timer::~Timer() {
  //  cout << " Timer::~Timer() " << Name() << endl;
}


void Timer::start(void) {

  ++count; 
  last_WCtime_stamp  = getTimeStampWC();

#ifdef PAPI
#ifdef NDEBUG
  PAPI_start(ID);
#else
  assert(PAPI_start(ID) == PAPI_OK); 
#endif
  //printf(" start ID%d\n", ID);
#endif  

#ifdef MEMORY_USAGE
  sprintf(stuff, "%30s start %10d  mem=%-10.3f\n", Name().c_str(), (long)Count(), get_memory_usage());
  pout() << stuff;
#endif
}

void Timer::stop(void) {

  accumulated_WCtime += getTimeStampWC() - last_WCtime_stamp;

#ifdef PAPI
#ifdef NDEBUG
  PAPI_stop(ID, values);
#else
  assert(PAPI_stop(ID, values) == PAPI_OK);
#endif
  accumulated_counter1 += values[0];
  accumulated_counter2 += values[1];
#endif  

#ifdef MEMORY_USAGE
  sprintf(stuff, "%30s stop  %10d  mem=%-10.3f\n", Name().c_str(), (long)Count(), get_memory_usage());
  pout() << stuff;
#endif

}

void Timer::clear(void) {

  accumulated_WCtime = 0.;
  count = 0;

#ifdef PAPI
  accumulated_counter1 = 0;
  accumulated_counter2 = 0;
#endif 

}

inline double Timer::getTimeStampWC(){

#ifdef MPI
  return( MPI_Wtime() );
#else    
  gettimeofday(&tv, &tz);
  //printf( "%d seconds, %d microseconds\n", tv.tv_sec, tv.tv_usec); 
  //cout << tvbuf->tv_sec << endl;
  //return(  ((double)((tvbuf->tv_sec)*1000000 + tvbuf->tv_usec))*1.0e-6 );

  // this doesn't always work cuz of potential integer overflow.
  //return( (tv.tv_sec * 1000000 + tv.tv_usec) * 1.0e-6 );

  // this is what MPICH uses for MPI_Wtime on machines
  // which have gettimeofday.
  return((double)tv.tv_sec + 0.000001 * (double)tv.tv_usec);
#endif
}


// This is the Summary for the Managed Timers.  If running in
// parallel, i first reduce all of the accumulated times for each
// timer and get the minimum, maximum, and average.  Then i go thru
// the list of timers and make a list of parent timers.  From there i
// can make for loops that step thru the tree and print out the
// results.
void Timer::TimerSummary(void) {

  // all ranks go in, but only rank0 writes the file.
  TimerSummary_();
  
  // all ranks need to clean up
#ifdef PAPI

  // not sure how important this cleanup and shutdown is...
  //for(int i=0; i<ID_counter; i++) {
  //printf(" cleaning up event %d\n", i);
  //PAPI_cleanup(&i);
  //}

  PAPI_shutdown();
#endif

  // clean up
  delete TimerList;
  
}

void Timer::TimerSummary_(void) {

  int rank, number_procs;
#ifdef MPI
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &number_procs);
#else
  rank=0;
  number_procs=1;
#endif

  list<Timer*>::iterator tli;  

  // don't write these files out all the darn time...
#ifdef WRITE_TITA_FILES
  FILE *TT;
  char ttfilename[80];
  if (number_procs < 999) {
    sprintf(ttfilename, "tita%03d", rank);
  } else {
    sprintf(ttfilename, "tita%04d", rank);
  }
  TT = fopen(ttfilename, "w");
  if(TT == NULL) {
    printf("problem opening tita\n");  
    exit(0); 
  }


#ifdef PAPI
  for(list<Timer*>::iterator tli=TimerList->begin() ; tli != TimerList->end() ; ++tli) {      
    double pc1 = (double)(*tli)->papi_counter1();
    double pc2 = (double)(*tli)->papi_counter2();
    double dc =  computeDerivedCounter(pc1, pc2);
    fprintf(TT, " %-16s P: %-16s d%-3d n=%14.0f wc=%13.3f pc1=%17.0f pc2=%17.0f dc=%13.3f\n",
            (*tli)->Name().c_str(), (*tli)->Parent.Name().c_str(),
            (*tli)->diagnostic_table, (double)((*tli)->Count()), (*tli)->wc_time(),
	    pc1, pc2, dc);
  }
#else
  for(list<Timer*>::iterator tli=TimerList->begin() ; tli != TimerList->end() ; ++tli) {      
    fprintf(TT, " %-16s P: %-16s d%-3d n=%14.0f  wc=%13.3f\n",
            (*tli)->Name().c_str(), (*tli)->Parent.Name().c_str(),
            (*tli)->diagnostic_table, (double)((*tli)->Count()), (*tli)->wc_time());
  }
#endif
  fclose(TT);
#endif

  //long long int totalCount;

  // go thru all Timers in static list and obtain the avg/min/max of
  // all processors onto rank0 -- obviously trivial for serial runs.
  for(tli=TimerList->begin() ; tli != TimerList->end() ; ++tli) {
      
    //printf(" Timer: %-16s Parent: %-16s diagnostic = %d  n=%d  wc=%10.3f\n",
    // (*tli)->Name().c_str(), (*tli)->Parent.Name().c_str(), 
    // (*tli)->diagnostic_table), (*tli)->Count(), (*tli)->wc_time();
      
    double wc = (*tli)->wc_time();
    long long int nc = (*tli)->Count();

    if (number_procs == 1) {
      (*tli)->avgWC = wc;
      (*tli)->minWC = wc;
      (*tli)->maxWC = wc;
      (*tli)->totalCount = nc;
      (*tli)->avgCount = (double)nc;
    } else {

#ifdef MPI
         
      // after here, rank0 will be the only guy with the right answers
      MPI_Reduce(&wc, &(*tli)->minWC, 1, MPI_DOUBLE, 
		 MPI_MIN, 0, MPI_COMM_WORLD);
      MPI_Reduce(&wc, &(*tli)->maxWC, 1, MPI_DOUBLE, 
		 MPI_MAX, 0, MPI_COMM_WORLD);

      double temp;
      MPI_Reduce(&wc, &temp, 1, MPI_DOUBLE, 
		 MPI_SUM, 0, MPI_COMM_WORLD);

      if(rank==0) (*tli)->avgWC = temp/number_procs;

      MPI_Reduce(&nc, &(*tli)->totalCount, 1, MPI_LONG_LONG, 
                 MPI_SUM, 0, MPI_COMM_WORLD);
      
      if(rank==0) {
        (*tli)->avgCount = (double)(*tli)->totalCount/number_procs;
      }
#endif

    }

    // serial or parallel.  assumes serial has rank set to 0.
    if(rank==0) {
      // in parallel, totalCount is the sum from all processors
      // but only exists on rank0.
      NtotalStartStops += (double)(*tli)->totalCount;
    }


      
  }



#ifdef PAPI
  // go thru all Timers in static list and obtain the sum of
  // PAPI counters onto rank0 -- obviously trivial for serial runs.
  for(tli=TimerList->begin() ; tli != TimerList->end() ; ++tli) {
      
    //printf(" Timer: %-16s Parent: %-16s diagnostic = %d  pc1=%14.5e pc2=%14.5e\n",
    ///     (*tli)->Name().c_str(), (*tli)->Parent.Name().c_str(), 
    //    (*tli)->diagnostic_table, 
    //     (double)(*tli)->papi_counter1(), 
    //     (double)(*tli)->papi_counter2());
      
    double pc1 = (double)(*tli)->papi_counter1();
    double pc2 = (double)(*tli)->papi_counter2();
      
    if (number_procs == 1) {
      (*tli)->totalPapiCounter1 = pc1;
      (*tli)->totalPapiCounter2 = pc2;
    } else {

#ifdef MPI
      MPI_Reduce(&pc1, &(*tli)->totalPapiCounter1, 1, MPI_DOUBLE, 
		 MPI_SUM, 0, MPI_COMM_WORLD);
      MPI_Reduce(&pc2, &(*tli)->totalPapiCounter2, 1, MPI_DOUBLE, 
		 MPI_SUM, 0, MPI_COMM_WORLD);
#endif
    } 

    //if (rank==0) {
    // printf(" Timer: %-16s after reduce pc1=%14.5e pc2=%14.5e  pc1=%14.5e pc2=%14.5e\n",
    //        (*tli)->Name().c_str(), 
    //        (*tli)->totalPapiCounter1, 
    //        (*tli)->totalPapiCounter2,
    //        (*tli)->total_papi_counter1(),
    //        (*tli)->total_papi_counter2());
    // }
  }
#endif  // PAPI


  // go away if you aren't rank 0
  if (rank != 0) { return; }
   
  // Create a list of Parent Timers from the list of Timers
  list<Timer*> ParentList;
  list<Timer*>::iterator pti; 

  //cout << " size of TimerList = " << TimerList.size() << endl;

  // for each timer in entire Timer List
  for(tli=TimerList->begin() ; tli != TimerList->end() ; tli++) {

    //printf(" Timer #%3d: %-16s Parent: %-16s diagnostic = %d\n",
    //   (*tli)->ID,
    //   (*tli)->Name().c_str(), 
    //   (*tli)->Parent.Name().c_str(), 
    //   (*tli)->diagnostic);

    // add the Parent of this Timer to the Parent List
    bool add = true;
      
    // but first, make sure it isn't already in the Parent List
    // for each parent currently in Parent List
    for(pti=ParentList.begin() ; pti != ParentList.end() ; pti++) {
      if(*pti == &((*tli)->Parent) ) {
	add = false;      
	break;
      }
    }
      
    if(add) {
      ParentList.push_back( &((*tli)->Parent) );
      //cout << "  Adding Parent:" << (*tli)->Parent.Name() << endl;
    }
    //cout << " ParentList size = " << ParentList.size() << endl;
  }


  // a hack to control the units. (minutes, seconds, etc)
  double time_unit_factor = 1.0;
  
  double largest_time = 0.0;
  for(pti=ParentList.begin() ; pti != ParentList.end() ; pti++) {
    double pa = (*pti)->avgWC*time_unit_factor;
    if (pa > largest_time) largest_time = pa; 
  }

  Timer TimerLoop;
  Timer iLooper0;
  const int Nloop = 10000;
  TimerLoop.start();
  for(int i=0; i<Nloop; i++) {
    iLooper0.start();    iLooper0.stop();
  }
  TimerLoop.stop();
  double TimerCost = TimerLoop.wc_time()/(double)iLooper0.Count();
  //printf(" %12.3e start/stop calls (%12.6f) in %12.6f seconds at %12.4f calls/s (%12.4e)\n",
  //     (double)iLooper0.Count(), iLooper0.wc_time(),
  //     TimerLoop.wc_time(), (double)iLooper0.Count()/TimerLoop.wc_time(),
  //     TimerCost);
  pout() << "Single Timer Start/Stop Cost = " << TimerCost << " sec." << endl;
  pout() << "Total Timer Start/Stop Calls = " << NtotalStartStops << endl;
  pout() << "Avg Timer Start/Stop Calls per Proc= " << NtotalStartStops/number_procs << endl;
  pout() << "Avg Estimated Timer Cost per Proc= " << NtotalStartStops*TimerCost/number_procs
         << " sec. (" << (NtotalStartStops*TimerCost/number_procs)/largest_time*100.0 
         << "%) " << endl;

   
  cout << " rank" << rank << " writing time.table " << endl;

  FILE *OUT;
  
  if(TIMER_COUNTER == 0) {
    OUT = fopen("time.table", "w");
  } else if (TIMER_COUNTER == 1) {
    OUT = fopen("time.table1", "w");
  } else {
    OUT = fopen("time.table2", "w");
  }
  if(OUT == NULL) {
    printf("problem opening output file in Timer\n");  
    exit(0); 
  }
  
  fprintf(OUT, "Number of Nodes: %d\n", number_procs );
  fprintf(OUT, "\n");
  
  for(pti=ParentList.begin() ; pti != ParentList.end() ; pti++) {
    
    fprintf(OUT, "\n");

    writeTableHeader(OUT);
      
    double table_avg_sum     = 0.0;
    double table_min_sum     = 0.0;
    double table_max_sum     = 0.0;
    double table_count_sum   = 0.0;
    double table_pc1_sum     = 0.0;
    double table_pc2_sum     = 0.0;
    double table_percent;

    double wc_avg = (*pti)->avgWC*time_unit_factor;
    double wc_min = (*pti)->minWC*time_unit_factor;
    double wc_max = (*pti)->maxWC*time_unit_factor;
    double count_avg = (*pti)->avgCount;

    double parent_avg = wc_avg;

    // Parent for each table
#ifdef PAPI
    //writeLineOfData(OUT, (*pti)->Name().c_str(), (*pti)->Count(), -1, 
    //	    wc_avg, wc_min, wc_max,
    //	    (*pti)->total_papi_counter1(), 
    //	    (*pti)->total_papi_counter2(),
    //              number_procs);
    writeLineOfData(OUT, (*pti)->Name().c_str(), count_avg, 100*wc_avg/largest_time, 
		    wc_avg, wc_min, wc_max,
		    (*pti)->total_papi_counter1(), 
		    (*pti)->total_papi_counter2(),
                    number_procs);
#else
    //writeLineOfData(OUT, (*pti)->Name().c_str(), (*pti)->Count(), -1,
    //	    wc_avg, wc_min, wc_max, 0, 0, 1);
    writeLineOfData(OUT, (*pti)->Name().c_str(), count_avg, 100*wc_avg/largest_time,
		    wc_avg, wc_min, wc_max, 0, 0, 1);
#endif

    writeSingleLineSeparator(OUT);

    for(tli=TimerList->begin() ; tli != TimerList->end() ; tli++) {
      
      if ((*tli)->diagnostic_table) continue;
      
      if( &((*tli)->Parent) == *pti && *tli != &((*tli)->Parent) ) {

  	wc_avg = (*tli)->avgWC*time_unit_factor;
  	wc_min = (*tli)->minWC*time_unit_factor;
  	wc_max = (*tli)->maxWC*time_unit_factor;
        count_avg = (*tli)->avgCount;

 	if(parent_avg  > 0.) {
 	  table_percent  = wc_avg/parent_avg*100.0;
 	} else {
 	  table_percent = 0.;
	}

  	table_avg_sum  += wc_avg;	
  	table_min_sum  += wc_min;	
  	table_max_sum  += wc_max;	
        //table_count_sum += (*tli)->Count();
        table_count_sum += count_avg;

#ifdef PAPI        
        table_pc1_sum += (*tli)->total_papi_counter1();
        table_pc2_sum += (*tli)->total_papi_counter2();

	// Children
	//writeLineOfData(OUT, (*tli)->Name().c_str(), (*tli)->Count(), 
        //	table_percent, wc_avg, wc_min, wc_max,
        //	(*tli)->total_papi_counter1(), 
        //	(*tli)->total_papi_counter2(),
        //              number_procs);	
	writeLineOfData(OUT, (*tli)->Name().c_str(), count_avg, 
			table_percent, wc_avg, wc_min, wc_max,
			(*tli)->total_papi_counter1(), 
			(*tli)->total_papi_counter2(),
                        number_procs);	
#else
	//writeLineOfData(OUT, (*tli)->Name().c_str(), (*tli)->Count(), 
        //	table_percent, wc_avg, wc_min, wc_max, 0, 0, 1);
	writeLineOfData(OUT, (*tli)->Name().c_str(), count_avg, 
			table_percent, wc_avg, wc_min, wc_max, 0, 0, 1);
#endif
      }
      
    } // end of table loop for this parent

    writeTableTotals(OUT, table_count_sum, parent_avg, 
                     table_avg_sum, table_min_sum,  table_max_sum,
                     table_pc1_sum, table_pc2_sum, TimerCost);
    
    
    fprintf(OUT,"\n\n");
  } // end of parent loop


  // Create a set of Diagnostic Tables -- ie all of the Timers
  // with unique values of member data "diagnostic_table".
  set<int> diagTableSet;

  for(tli=TimerList->begin() ; tli != TimerList->end() ; tli++) {    
    if ((*tli)->diagnostic_table) {
      diagTableSet.insert((*tli)->diagnostic_table);
    }
  }

  //cout << " Number of Diag Tables = " << diagTableSet.size() << endl;
  //for (set<int>::iterator si=diagTableSet.begin(); si != diagTableSet.end(); si++) {
  // cout << " si = " << *si << endl;
  //}
  
  // For each unique value of diagnostic_table, print a table.
  for (set<int>::iterator si=diagTableSet.begin(); si != diagTableSet.end(); si++) {

    fprintf(OUT, "\n\n Diagnostic Table %d\n", *si);

    // crude way of getting the Parent for this diagnostic table.
    // part of the crudeness is that i'm not checking that all
    // Timers in this diagnostic table have the same parent.
    Timer *DiagnosticParent = NULL;
    for(tli=TimerList->begin() ; tli != TimerList->end() ; tli++) {      
      if ((*tli)->diagnostic_table != *si) continue;
      DiagnosticParent = &((*tli)->Parent);
    }
    
    writeTableHeader(OUT);
    
    double table_avg_sum   = 0.0;
    double table_min_sum   = 0.0;
    double table_max_sum   = 0.0;
    double table_count_sum = 0.0;
    double table_pc1_sum   = 0.0;
    double table_pc2_sum   = 0.0;
    double table_percent;
    
    double wc_avg = DiagnosticParent->avgWC*time_unit_factor;
    double wc_min = DiagnosticParent->minWC*time_unit_factor;
    double wc_max = DiagnosticParent->maxWC*time_unit_factor;
    double count_avg = DiagnosticParent->avgCount;

    double parent_avg = wc_avg;
    
    // Parent for each table
#ifdef PAPI
    //writeLineOfData(OUT, DiagnosticParent->Name().c_str(), 
    //	    DiagnosticParent->Count(), -1,
    //	    wc_avg, wc_min, wc_max,
    //	    DiagnosticParent->total_papi_counter1(), 
    //	    DiagnosticParent->total_papi_counter2(),
    //              number_procs);
    writeLineOfData(OUT, DiagnosticParent->Name().c_str(), 
		    count_avg, -1,
		    wc_avg, wc_min, wc_max,
		    DiagnosticParent->total_papi_counter1(), 
		    DiagnosticParent->total_papi_counter2(),
                    number_procs);
#else
    //writeLineOfData(OUT, DiagnosticParent->Name().c_str(), 
    //	    DiagnosticParent->Count(), -1, 
    //	    wc_avg, wc_min, wc_max, 0, 0, 1);
    writeLineOfData(OUT, DiagnosticParent->Name().c_str(), 
		    count_avg, -1, 
		    wc_avg, wc_min, wc_max, 0, 0, 1);
#endif
    
    writeSingleLineSeparator(OUT);  
    
    for(tli=TimerList->begin() ; tli != TimerList->end() ; tli++) {
      
      // skip all Timers except ones belonging in this table
      if ((*tli)->diagnostic_table != *si) continue;
      
      wc_avg = (*tli)->avgWC*time_unit_factor;
      wc_min = (*tli)->minWC*time_unit_factor;
      wc_max = (*tli)->maxWC*time_unit_factor;
      count_avg = (*tli)->avgCount;

      if(parent_avg  > 0.) {
        table_percent  = wc_avg/parent_avg*100.0;
      } else {
        table_percent = 0.;
      }
      
      table_avg_sum  += wc_avg;	
      table_min_sum  += wc_min;	
      table_max_sum  += wc_max;	
      //table_count_sum += (*tli)->Count();
      table_count_sum += count_avg;

#ifdef PAPI
      table_pc1_sum += (*tli)->total_papi_counter1();
      table_pc2_sum += (*tli)->total_papi_counter2();
      
      // Children
      //writeLineOfData(OUT, (*tli)->Name().c_str(), (*tli)->Count(), 
      //              table_percent, wc_avg, wc_min, wc_max,
      //              (*tli)->total_papi_counter1(), 
      //              (*tli)->total_papi_counter2(),
      //              number_procs);
      writeLineOfData(OUT, (*tli)->Name().c_str(), count_avg, 
                      table_percent, wc_avg, wc_min, wc_max,
                      (*tli)->total_papi_counter1(), 
                      (*tli)->total_papi_counter2(),
                      number_procs);
      
#else
      //writeLineOfData(OUT, (*tli)->Name().c_str(), (*tli)->Count(), 
      //              table_percent, wc_avg, wc_min, wc_max, 0, 0, 1);
      writeLineOfData(OUT, (*tli)->Name().c_str(), count_avg, 
                      table_percent, wc_avg, wc_min, wc_max, 0, 0, 1);
#endif
      
    } // end of table loop for this parent    
    
    writeTableTotals(OUT, table_count_sum, parent_avg, 
                     table_avg_sum, table_min_sum,  table_max_sum,
                     table_pc1_sum, table_pc2_sum, TimerCost);
    
    fprintf(OUT,"\n\n");
  }


  fprintf(OUT, "Single Timer Start/Stop Cost = %15.7e sec.\n", TimerCost);
  fprintf(OUT, "Total Timer Start/Stop Calls = %20.10e \n", NtotalStartStops);
  fprintf(OUT, "Avg Timer Start/Stop per Proc= %20.10e \n", NtotalStartStops/number_procs);
  fprintf(OUT, "Avg Est Timer Cost per Proc  = %15.3f sec.  (%7.2f%%)\n", 
          NtotalStartStops*TimerCost/number_procs, 
          (NtotalStartStops*TimerCost/number_procs)/largest_time*100.0);

  fclose(OUT);
 
  //FILE *QOUT;
  //char qfilename[80];
  //sprintf(qfilename, "q%03d.dat", rank);
  //QOUT = fopen(qfilename, "w");
  //if(QOUT == NULL) {
  //printf("problem opening pex out\n");  
  //exit(0); 
  //}

  //  for(unsigned int i=0; i<NQ; i++) {
  //fprintf(OUT,"%20.10e  s0=%20.10e s1=%20.10e s2=%20.10e\n",
  //        (double)i, (double)qsubbox_len0[i], 
  //        (double)qsubbox_len1[i], (double)qsubbox_len2[i]);
  //}
  //fprintf(OUT,"iter= %20.10e N=%20.10e\n",
  //      (double)qiter, (double)qNN);

  //for(unsigned int i=0; i<qiter; i++) {
  //fprintf(QOUT, "%12d %4d %4d %4d %4d\n",
  //        i, qNN[i], qsubbox_len2[i], qsubbox_len1[i], qsubbox_len0[i]);
  //}
  //fclose(QOUT);

}


static void writeSingleLineSeparator(FILE *out) {
#ifdef PAPI  
  fprintf(out, "-------------------------------------------------");
  fprintf(out, "-------------------------------------------------\n");
#else
  fprintf(out, "-------------------------------------------------");
  fprintf(out, "---------------------------\n");
#endif
}

static void writeDoubleLineSeparator(FILE *out) {
#ifdef PAPI
  fprintf(out, "=================================================");
  fprintf(out, "=================================================\n");
#else
  fprintf(out, "=================================================");
  fprintf(out, "===========================\n");
#endif
}


static void writeTableHeader(FILE *out) {


#ifdef PAPI
  ////fprintf(out, "                                       ");
  ////fprintf(out, "wall-clock time (sec)\n");
  ////fprintf(out, "                                      ");
  ////fprintf(out, "_______________________\n");
  
  //            12345678901234567890
  fprintf(out, "      Totals    ");
  fprintf(out, "  WC per.");
  fprintf(out, "      count");

  fprintf(out, "    avg  ");
  fprintf(out, "     min  ");
  fprintf(out, "     max  ");

  if(TIMER_COUNTER==0) {
    fprintf(out, "    cycles     FP INS     mflops");
  } else if(TIMER_COUNTER==1) {
    fprintf(out, "    L1 TCM     L2 TCM     L2/L1");
  } else {
    fprintf(out, "    TOT INS    BR INS     BR/TOT");
  }
#else
  ////fprintf(out, "                                             ");
  ////fprintf(out, "wall-clock time (sec)\n");
  ////fprintf(out, "                                         ");
  ////fprintf(out, "______________________________\n");
  
  //            12345678901234567890
  fprintf(out, "      Totals    ");
  fprintf(out, "  WC per.");
  fprintf(out, "      count");

  fprintf(out, "        avg  ");
  fprintf(out, "       min  ");
  fprintf(out, "       max  ");
#endif

  fprintf(out, "\n");

  writeDoubleLineSeparator(out);  

}

static void writeTableTotals(FILE *out, 
                             const double count_sum,
			     const double parent_avg,
			     const double avg_sum, 
			     const double min_sum,
			     const double max_sum,
                             const double pc1_sum,
                             const double pc2_sum,
                             const double TimerCost) {

//   const char tcs[16] = "Timer Cost";

//   double totalTimerCost = (double)count_sum * TimerCost;
//   double percentTimerCost;
//   if(parent_avg  > 0.) {
//     percentTimerCost = totalTimerCost/parent_avg * 100.0;
//   } else {
//     percentTimerCost = 0.;
//   }

// #ifdef PAPI
//   fprintf(out, "%16s (%6.2f%%) %9ld %8.2f [%8.2f,%8.2f]  %10.3e %10.3e %8.2f\n",
//           tcs, percentTimerCost, (long)count_sum, totalTimerCost, 0.0, 0.0, 0.0, 0.0, 0.0);
// #else
//   fprintf(out, "%16s (%6.2f%%) %9ld %10.2f [%10.2f, %10.2f]\n",
//           tcs, percentTimerCost, (long)count_sum, totalTimerCost, 0.0, 0.0);
// #endif

  writeDoubleLineSeparator(out);

  double percent;
  if(parent_avg  > 0.) {
    percent = avg_sum/parent_avg * 100.0;
  } else {
    percent = 0.;
  }

#ifdef PAPI
  double derived_counter = computeDerivedCounter(pc1_sum, pc2_sum);

  fprintf(out, "  table totals:   %6.2f%%  %9ld %8.2f [%8.2f,%8.2f]  %10.3e %10.3e %8.2f\n",
	  percent, (long)count_sum, 
	  avg_sum, min_sum, max_sum,
          pc1_sum, pc2_sum, derived_counter); 
#else
  fprintf(out, "  table totals:   %6.2f%%  %9ld %10.2f [%10.2f, %10.2f]\n", 
	  percent, (long)count_sum, 
	  avg_sum, min_sum, max_sum); 
#endif

}

// static void writeLineOfData(FILE *out,
// 			    const char *name,
// 			    const long int count,
// 			    const double percent,
// 			    const double wc_avg,
// 			    const double wc_min,
// 			    const double wc_max,
// 			    const double counter1,
// 			    const double counter2,
//                             const int number_procs) {

// #ifdef PAPI

//   // PAPI counters should be coming in as sum totals of all
//   //  processors. 

//   double derived_counter = computeDerivedCounter(counter1, counter2);

//   if(percent < 0) {
//     fprintf(out, "%16s           %9ld %8.2f [%8.2f,%8.2f]  %10.3e %10.3e %8.2f\n",
// 	    name, count, wc_avg, wc_min, wc_max,
// 	    counter1, counter2, 
// 	    derived_counter);
//   } else {
//     fprintf(out, "%16s (%6.2f%%) %9ld %8.2f [%8.2f,%8.2f]  %10.3e %10.3e %8.2f\n",
// 	    name, percent, count, wc_avg, wc_min, wc_max,
// 	    counter1, counter2, 
// 	    derived_counter);
//   }
// #else
//   if(percent < 0) {
//     fprintf(out, "%16s           %9ld %10.2f [%10.2f, %10.2f]\n",
// 	    name, count, wc_avg, wc_min, wc_max);  
//   } else {
//     fprintf(out, "%16s (%6.2f%%) %9ld %10.2f [%10.2f, %10.2f]\n",
// 	    name, percent, count, wc_avg, wc_min, wc_max);  
//   }
// #endif
  
// }

static void writeLineOfData(FILE *out,
			    const char *name,
			    const double count,
			    const double percent,
			    const double wc_avg,
			    const double wc_min,
			    const double wc_max,
			    const double counter1,
			    const double counter2,
                            const int number_procs) {

#ifdef PAPI

  // PAPI counters should be coming in as sum totals of all
  //  processors. 

  double derived_counter = computeDerivedCounter(counter1, counter2);

  if(percent < 0) {
    fprintf(out, "%16s           %9.0f %8.2f [%8.2f,%8.2f]  %10.3e %10.3e %8.2f\n",
	    name, count, wc_avg, wc_min, wc_max,
	    counter1, counter2, 
	    derived_counter);
  } else {
    fprintf(out, "%16s  %6.2f%%  %9.0f %8.2f [%8.2f,%8.2f]  %10.3e %10.3e %8.2f\n",
	    name, percent, count, wc_avg, wc_min, wc_max,
	    counter1, counter2, 
	    derived_counter);
  }
#else
  if(percent < 0) {
    fprintf(out, "%16s           %9.0f %10.2f [%10.2f, %10.2f]\n",
	    name, count, wc_avg, wc_min, wc_max);  
  } else {
    fprintf(out, "%16s  %6.2f%%  %9.0f %10.2f [%10.2f, %10.2f]\n",
	    name, percent, count, wc_avg, wc_min, wc_max);  
  }
#endif
  
}

#ifdef PAPI

double computeDerivedCounter(double pc1, double pc2) {

  const double VALID_PROC_TIME = 1.0e-4;

  double proc_time = 1.0;
  double derived_counter = 0.0;

  if(TIMER_COUNTER == 0) {
    if(CPU_MHZ <= 0.0) CPU_MHZ = 1.0;
    proc_time = pc1/(CPU_MHZ*1.0e6);
    if(proc_time > VALID_PROC_TIME) {
      derived_counter = pc2/(proc_time*1.0e6);
    } else {
      derived_counter = 0.0;
    }
  } else if (TIMER_COUNTER == 1) {    
    if(pc1 > 0) {
      derived_counter = pc2/pc1;
    } else {
      derived_counter = 0.0;
    }
  } else {
    if(pc1 > 0) {
      derived_counter = pc2/pc1;
    } else {
      derived_counter = 0.0;
    }
  }

  return derived_counter;
}

static int PAPI_initSingleEvent(const int event, int *tag) {

  int retval;
  retval = PAPI_query_event(event);
  assert(retval == PAPI_OK);

  retval = PAPI_create_eventset(tag);
  assert(retval == PAPI_OK);

  retval = PAPI_add_event(tag, event);
  assert(retval == PAPI_OK);
  return retval; // to get rid of warning mesg
}

static int PAPI_initDualEvent(const int event1, const int event2, 
			       int *tag) {
  int retval;

  retval = PAPI_query_event(event1);
  assert(retval == PAPI_OK);

  retval = PAPI_query_event(event2);
  assert(retval == PAPI_OK);

  retval = PAPI_create_eventset(tag);
  assert(retval == PAPI_OK);

  retval = PAPI_add_event(tag, event1);
  assert(retval == PAPI_OK);

  retval = PAPI_add_event(tag, event2);
  assert(retval == PAPI_OK);
  return retval;  // to get rid of warning mesg
}


static int PAPI_initMultiEvent(const int event1, const int event2, 
				const int event3, int *tag) {

  int retval;

  retval = PAPI_query_event(event1);
  assert(retval == PAPI_OK);
  retval = PAPI_query_event(event2);
  assert(retval == PAPI_OK);
  retval = PAPI_query_event(event3);
  assert(retval == PAPI_OK);

  retval = PAPI_create_eventset(tag);
  assert(retval == PAPI_OK);

  retval = PAPI_add_event(tag, event1);
  assert(retval == PAPI_OK);
  retval = PAPI_add_event(tag, event2);
  assert(retval == PAPI_OK);

  retval = PAPI_set_multiplex(tag);
  assert(retval == PAPI_OK);

  retval = PAPI_add_event(tag, event3);
  assert(retval == PAPI_OK);
  return retval; // to get rid of warning mesg
}


static int PAPI_checkEvent(const int event) {
  int tag = PAPI_NULL;
  long long int value = -1;
  int retval;

  if( (retval = PAPI_query_event(event)) != PAPI_OK ) {
    printf(" problem with PAPI_query_event retval=%d\n", retval);    
    return retval;
  }
  if( (PAPI_create_eventset(&tag) != PAPI_OK) ) {
    printf(" problem with PAPI_create_eventset\n");
    return retval;
  }
  
  if( (PAPI_add_event(&tag, event) != PAPI_OK) ) {
    printf(" problem with PAPI_add_event\n");
    return retval;
  }
  if( (PAPI_start(tag) != PAPI_OK) ) {
    printf(" problem with PAPI_start\n");
    return retval;
  }
  if( (PAPI_stop(tag, &value) != PAPI_OK) ) {
    printf(" problem with PAPI_stop\n");
    return retval;
  }
  printf(" event looks ok.  value=%lld\n", value);
  return retval; // to get rid of warning mesg
}



static int PAPI_init(void) {
  int retval;
  retval = PAPI_library_init(PAPI_VER_CURRENT);
  assert(retval == PAPI_VER_CURRENT);
  return retval; // to get rid of warning mesg
}

static void PAPI_hwinfo(float &mhz) {
  const PAPI_hw_info_t *hwinfo = NULL;
  hwinfo = PAPI_get_hardware_info();
  assert( hwinfo != NULL );

  printf("Vendor string and code   : %s (%d)\n",hwinfo->vendor_string,hwinfo->vendor);
  printf("Model string and code    : %s (%d)\n",hwinfo->model_string,hwinfo->model);
  printf("CPU revision             : %f\n",hwinfo->revision);
  printf("CPU Megahertz            : %f\n",hwinfo->mhz);
  printf("CPU's in an SMP node     : %d\n",hwinfo->ncpu);
  printf("Nodes in the system      : %d\n",hwinfo->nnodes);
  printf("Total CPU's in the system: %d\n",hwinfo->totalcpus);

  mhz = hwinfo->mhz;
}

static void PAPI_initMultiplex(void) {
  int retval = PAPI_multiplex_init();
  assert(retval == PAPI_OK);
}

static void PAPI_cleanup(int *i) {
#ifdef NDEBUG  // no debug
  PAPI_cleanup_eventset(i);   
#else
  assert(PAPI_cleanup_eventset(i) == PAPI_OK);   
#endif
}

#endif // on PAPI

#else
#include "Timer.H"

#endif  // TIMER
